#include <ble.h>
#include <sd_rpc.h>

#include <cassert>
#include <vector>

#include <iostream>
#include <unistd.h> // sleep

static const uint32_t config_id = 1;

static void check_error( uint32_t error_code )
{
    static_cast< void >( error_code );
    assert( error_code == NRF_SUCCESS );
}

static constexpr uint16_t scan_interval = 0x00A0; /**< Determines scan interval in units of 0.625 milliseconds. */
static constexpr uint16_t scan_window   = 0x0050; /**< Determines scan window in units of 0.625 milliseconds. */
static constexpr uint16_t scan_timeout  = 0x0;    /**< Scan timeout between 0x01 and 0xFFFF in seconds, 0x0 disables timeout. */

static uint8_t     report_data[100]     = { 0 };

extern "C" {

    static void start_async_scan_impl( adapter_t* that )
    {
        static ble_data_t  adv_report_buffer;

        adv_report_buffer.p_data = report_data;
        adv_report_buffer.len = sizeof( report_data );

        static const ble_gap_scan_params_t scan_params =
        {
            0,                       // Set if accept extended advertising packetets.
            0,                       // Set if report inomplete reports.
            1,                       // Set if active scanning.
            BLE_GAP_SCAN_FP_ACCEPT_ALL,
            BLE_GAP_PHY_1MBPS,
            scan_interval,
            scan_window,
            scan_timeout
            , { 0 }                  // Set channel mask.
        };

        check_error( sd_ble_gap_scan_start( that, &scan_params, &adv_report_buffer ) );
    }
}

static adapter_t *adapter_init(const char* serial_port, std::uint32_t baud_rate )
{
    physical_layer_t* phy = sd_rpc_physical_layer_create_uart(serial_port,
                                            baud_rate,
                                            SD_RPC_FLOW_CONTROL_NONE,
                                            SD_RPC_PARITY_NONE);
    data_link_layer_t* data_link_layer = sd_rpc_data_link_layer_create_bt_three_wire(phy, 250);
    transport_layer_t* transport_layer = sd_rpc_transport_layer_create(data_link_layer, 1500);

    return sd_rpc_adapter_create(transport_layer);
}

static void status_handler(adapter_t * adapter, sd_rpc_app_status_t code, const char * message)
{
    (void)adapter;
    printf("Status: %d, message: %s\n", (uint32_t)code, message);
    fflush(stdout);
}

static void on_adv_report( adapter_t * adapter, const ble_gap_evt_adv_report_t& adv_report )
{
    // Log the Bluetooth device address of advertisement packet received.
    // std::cout << "Received advertisement report with device address: "
    //     << address_from_nrf( adv_report.peer_addr ) << std::endl;
    const bool is_scan_response = adv_report.type.scan_response;


  std::cout << "connectable   : " << adv_report.type.connectable     << std::endl;
  std::cout << "scannable     : " << adv_report.type.scannable << std::endl;
  std::cout << "directed      : " << adv_report.type.directed << std::endl;
  std::cout << "scan_response : " << adv_report.type.scan_response << std::endl;
  std::cout << "extended_pdu  : " << adv_report.type.extended_pdu << std::endl;
  std::cout << "status        : " << adv_report.type.status << std::endl;
  std::cout << "reserved      : " << adv_report.type.reserved << std::endl;

std::cout << "is_scan_response: " << is_scan_response << std::endl;

    static ble_data_t  adv_report_buffer;

    adv_report_buffer.p_data = report_data;
    adv_report_buffer.len = sizeof( report_data );

    sd_ble_gap_scan_start( adapter, nullptr, &adv_report_buffer );
}

static void ble_evt_dispatch(adapter_t * adapter, ble_evt_t * p_ble_evt)
{
    assert( adapter );
    assert( p_ble_evt );

    switch (p_ble_evt->header.evt_id)
    {
        case BLE_GAP_EVT_CONNECTED:
            break;

        case BLE_GAP_EVT_DISCONNECTED:
            break;

        case BLE_GAP_EVT_ADV_REPORT:
            on_adv_report( adapter, p_ble_evt->evt.gap_evt.params.adv_report );
            break;

        case BLE_GAP_EVT_TIMEOUT:
            break;

        case BLE_GATTC_EVT_PRIM_SRVC_DISC_RSP:
            break;

        case BLE_GATTC_EVT_CHAR_DISC_RSP:
            break;

        case BLE_GATTC_EVT_DESC_DISC_RSP:
            break;

        case BLE_GATTC_EVT_WRITE_RSP:
            break;

        case BLE_GATTC_EVT_HVX:
            break;

        case BLE_GAP_EVT_CONN_PARAM_UPDATE_REQUEST:
            break;

        case BLE_GATTS_EVT_EXCHANGE_MTU_REQUEST:
            break;

        case BLE_GATTC_EVT_EXCHANGE_MTU_RSP:
            break;

        default:
            printf("Received an un-handled event with ID: %d\n", p_ble_evt->header.evt_id);
            fflush(stdout);
            break;
    }
}

static void log_handler(adapter_t * adapter, sd_rpc_log_severity_t severity, const char * message)
{
    (void)adapter;
    switch (severity)
    {
        case SD_RPC_LOG_ERROR:
            printf("Error: %s\n", message);
            fflush(stdout);
            break;

        case SD_RPC_LOG_WARNING:
            printf("Warning: %s\n", message);
            fflush(stdout);
            break;

        case SD_RPC_LOG_INFO:
            printf("Info: %s\n", message);
            fflush(stdout);
            break;

        default:
            printf("Log: %s\n", message);
            fflush(stdout);
            break;
    }
}

static void ble_cfg_set( adapter_t* adapter, std::uint8_t conn_cfg_tag )
{
    const std::uint32_t ram_start = 0; // Value is not used by ble-driver
    ble_cfg_t ble_cfg;

    // Configure the connection roles.
    memset( &ble_cfg, 0, sizeof(ble_cfg) );

    ble_cfg.gap_cfg.role_count_cfg.adv_set_count        = BLE_GAP_ADV_SET_COUNT_DEFAULT;
    ble_cfg.gap_cfg.role_count_cfg.periph_role_count    = 0;
    ble_cfg.gap_cfg.role_count_cfg.central_role_count   = 1;
    ble_cfg.gap_cfg.role_count_cfg.central_sec_count    = 0;

    check_error( sd_ble_cfg_set( adapter, BLE_GAP_CFG_ROLE_COUNT, &ble_cfg, ram_start ) );

    // Configure the MTU
    memset( &ble_cfg, 0, sizeof(ble_cfg) );
    ble_cfg.conn_cfg.conn_cfg_tag                 = conn_cfg_tag;
    ble_cfg.conn_cfg.params.gatt_conn_cfg.att_mtu = 150;

    check_error( sd_ble_cfg_set( adapter, BLE_CONN_CFG_GATT, &ble_cfg, ram_start ) );
}

int main()
{
    std::uint32_t *  app_ram_base = nullptr;
    const auto       nrf_adapter = adapter_init( "/dev/tty.usbmodemDCC9832470962", 1000000 );

    check_error( sd_rpc_open( nrf_adapter, status_handler, ble_evt_dispatch, log_handler) );
    ble_cfg_set( nrf_adapter, config_id );
    check_error( sd_ble_enable( nrf_adapter, app_ram_base ) );

    start_async_scan_impl( nrf_adapter );

    sleep(20);
}